package net.sf.corn.gate;

import java.io.Serializable;

/**
 * Defines a Scoped Data Area for service implementations, so that the service can
 * share data with the others.
 * 
 * @author Serhat Dirik
 */
public interface IDataStage extends Serializable{
	/**
	 * Defines the scope of a data stage
	 * @author Serhat Dirik
	 */
	public enum Scope {
		/**
		 * Data area only lives until service request is responded.
		 * This type data area is created when a new request come to the server.
		 */
		REQUEST, 
		/**
		 * Data area is bounded to a client session which managed by the platform.
		 */
		SESSION, 
		/**
		 * Application scoped Data Stage, some platforms like JEE WEB Container May Host Multiple Apps At the same platform 
		 */
		APPLICATION,
		/**
		 * Accessible by any class which lives in the same platform. Also known as  JVM Level data area.
		 */
		PLATFORM;
	}
	
	/**
	 * @return The scope of this data area
	 */
	public Scope getScope();
	
	/**
	 * Unique data area identifier
	 * @return a unique data stage identifier
	 */
	public String getId();
	
	/**
	 * Gives the attribute value that saved in the data stage previously
	 * @param name name of the attribute
	 * @return null if there is no attribute saved previously
	 */
	public Object getAttribute(String name);
	
    /**
     * Gives all attributes that are saved previously
     * @return saved attributes in the data stage
     */
	public Iterable<String> getAttributeNames();


	/**
	 * Removes an attribute from the data stage
	 * @param name name of the attribute
	 */
	public void removeAttribute(String name);
	
	/**
	 * Tells if the data stage contains a value that mapped with the given name
	 * @param name name of the attribute
	 * @return true if exist 
	 */
	public boolean hasAttribute(String name);
	
	/**
	 * Saves an attribute in to the data stage
	 * @param name name of the attribute
	 * @param value value of the attribute. In order to support session serialization, save only serializable objects
	 */
	public void setAttribute(String name, Object value);
}
