package net.sf.corn.gate;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.locks.ReentrantLock;

import net.sf.corn.cps.CPScanner;
import net.sf.corn.cps.ClassFilter;
import net.sf.corn.cps.CombinedFilter;
import net.sf.corn.cps.ResourceFilter;
import net.sf.corn.exception.ExceptionBuilder;
import net.sf.corn.exception.ExceptionCatcher;
import net.sf.corn.exception.ExceptionCommand;
import net.sf.corn.gate.IDataStage.Scope;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Registry keeps records of all important objects in the gate such as services,
 * filters, locator. Any service that need to be introduced to gate must be registered 
 * by using one of methods that registry offering.   
 * 
 * 
 * @author Serhat Dirik
 * 
 */
public class Registry {
	private static Logger log = LoggerFactory.getLogger(Registry.class);
	private static ReentrantLock lock = new ReentrantLock();
	public static final String SESSION_KEY_PREFIX = "JVSComponentInstance_";

	private static ConcurrentHashMap<String, ILocator> locators = new ConcurrentHashMap<>();
	private static CopyOnWriteArrayList<ServiceFilter> filters = new CopyOnWriteArrayList<>();
	private static ConcurrentHashMap<String, Object> componentSingletons = new ConcurrentHashMap<>();
	private static ConcurrentHashMap<String, ServiceDefinition> serviceDefinitions = new ConcurrentHashMap<>();
	private static ConcurrentHashMap<String, ReentrantLock> serviceLock = new ConcurrentHashMap<>();
	private static Class<? extends WrapperClientContext> contextWrapper = null;

	private static ExceptionCommand cmdLogAndRelease = new ExceptionCommand() {
		@Override
		public void execute(Exception e) {
			log.error(e.getMessage(), e);
			releaseState();
		}
	};

	private static ExceptionCommand cmdLogDebug = new ExceptionCommand() {
		public void execute(Exception e) {
			if (log.isDebugEnabled())
				log.debug(e.getMessage(), e);
		}
	};

	private static ExceptionCommand cmdLogError = new ExceptionCommand() {
		public void execute(Exception e) {
			if (log.isErrorEnabled())
				log.error(e.getMessage(), e);
		}
	};

	static {
		// Add default Locator
		try {
			registerLocator(POJOLocator.class.getCanonicalName());
		} catch (GateException e) {
			ExceptionBuilder
					.with(RuntimeException.class)
					.cause(e)
					.message("Exception occured while registering the defaults")
					.execute(cmdLogError).throwIt();
		}
	}

	/*
	 * Filters Related Ops
	 */
	
	/**
	 * Registers a filter
	 * 
	 * @param filterClass {@link IServiceFilter} implementation 
	 * @return true if the given filter is added
	 */
	public static boolean registerFilter(Class<? extends IServiceFilter> filterClass) {
		nullArgumentsCheck(filterClass);
		ServiceFilter filter = null;
		try {
			filter = (ServiceFilter) filterClass.newInstance();
		} catch (Exception e) {
			 
		}
		return registerFilter(filter);
	}
	
	/**
	 * Registers a filter
	 * 
	 * @param filter the filter to be registered 
	 * @return true if the given filter is added
	 */
	public static boolean registerFilter(ServiceFilter filter) {
		nullArgumentsCheck(filter);
		boolean ret = false;
		lockState();
		if (!filters.contains(filter)) {
			ret = filters.add(filter);
			if(log.isInfoEnabled()){
				log.info(filter.toString()+" filter is registered!");
			}
		}else{
			if(log.isWarnEnabled()){
				log.warn(filter.toString()+" filter is already registered!");
			}
		}
		releaseState();
		return ret;
	}

	/**
	 * Registers an advice implementation to the registry by putting it into a
	 * random named filter.The advice implementation must be annotated with
	 * {@link GateServiceTarget} or {@link GateServiceTargets} annotation.
	 * 
	 * @param advice  the advice implementation class
	 * @return true if it's registered successfully
	 * @throws GateException
	 */
	public static boolean registerAdvice(IExecutionAdvice advice)
			throws GateException {
		nullArgumentsCheck(advice);
		if (!advice.getClass().isAnnotationPresent(GateServiceTargets.class)
				&& !advice.getClass().isAnnotationPresent(
						GateServiceTarget.class)) {
			throw ExceptionBuilder
					.with(GateException.class)
					.message(
							advice.getClass().getName()
									+ " advice implementation is not marked with @GateServiceTarget or @GateServiceTargets annotation!")
					.build().code(GateException.CODE_REGISTRY_NOT_WELL_DEFINED);
		}
		// Build a random named filter
		ServiceFilter filter = new ServiceFilter(UUID.randomUUID().toString());
		// Attach targets
		GateServiceTargets targets = advice.getClass().getAnnotation(
				GateServiceTargets.class);
		if (targets != null) {
			for (GateServiceTarget target : targets.serviceTargets()) {
				ServiceTargetDefinition tdef = new ServiceTargetDefinition(
						target.servicePattern());
				for (String opPattern : target.operationPatterns()) {
					tdef.addOperationTarget(opPattern);
				}
				filter.addServiceTarget(tdef);
			}
		}
		// Attach target
		GateServiceTarget target = advice.getClass().getAnnotation(
				GateServiceTarget.class);
		if (target != null) {
			ServiceTargetDefinition tdef = new ServiceTargetDefinition(
					target.servicePattern());
			for (String opPattern : target.operationPatterns()) {
				tdef.addOperationTarget(opPattern);
			}
			filter.addServiceTarget(tdef);
		}

		if (filter.getServiceTargetDefinitions().size() == 0) {
			throw ExceptionBuilder
					.with(GateException.class)
					.message(
							advice.getClass().getName()
									+ " advice targets are not well defined!").build()
					.code(GateException.CODE_REGISTRY_NOT_WELL_DEFINED);
		}

		filter.addAdvice(advice);
		return registerFilter(filter);
	}

	 

	/**
	 * Unregisters the given filter
	 * 
	 * @param filter
	 *            the filter to be removed
	 */
	public static void unregisterFilter(ServiceFilter filter) {
		if (filter == null)
			return;
		lockState();
		filters.remove(filter);
		if(log.isInfoEnabled()){
			log.info(filter+" filter is unregistered");
		}
		releaseState();
	}

	/**
	 * Gives all registered filters as unmodifiable list
	 * 
	 * @return list of filters
	 */
	public static List<ServiceFilter> getFilters() {
		return Collections.unmodifiableList(filters);
	}

	/**
	 * Finds all registered filters for the given service name
	 * 
	 * @param serviceName
	 *            The name of the service
	 * @return list of filters
	 */
	public static List<ServiceFilter> findFilters(String serviceName) {
		return findFilters(serviceName, null);
	}

	/**
	 * Finds all filters for the given service name and the operation name
	 * 
	 * @param serviceName
	 *            The name of the service
	 * @param operationName
	 *            The name of the operation
	 * @return list of filters
	 */
	public static List<ServiceFilter> findFilters(String serviceName,
			String operationName) {
		List<ServiceFilter> retLst = new ArrayList<ServiceFilter>();
		for (ServiceFilter filter : filters) {
			if (filter.matches(serviceName, operationName)) {
				retLst.add(filter);
			}
		}
		return retLst;
	}

	/*
	 * 
	 */
	/**
	 * Context wrapping is a technique to manipulate context values. Consider a
	 * wep application that uses it's own security implementation instead of JEE
	 * standards. Such application needs to override context values such as user
	 * information to offer proper values to services. In that case wep app
	 * developers can write a {@link WrapperClientContext} extending class and
	 * register it by using this method to the registry. Thus, wrapper context
	 * wraps the default web context and override some values as desired.
	 * 
	 * @param className
	 *            A Class Name that extends {@link WrapperClientContext} class
	 */

	@SuppressWarnings("unchecked")
	public static void registerContextWrapper(String className)
			throws GateException {
		nullArgumentsCheck(className);
		if (contextWrapper != null && log.isWarnEnabled()) {
			log.warn(className
					+ " wrapper context will override the previous value "
					+ contextWrapper.getClass().getName());
		}
		// Create instance
		Class<? extends WrapperClientContext> contextClass = null;
		try {
			contextClass = (Class<? extends WrapperClientContext>) Class
					.forName(className);
		} catch (Exception e) {
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogError)
					.message("Context Wrapper " + className + " is not found !")
					.build().code(GateException.CODE_GENERIC_CLASS_NOT_FOUND);
		}

		if (!WrapperClientContext.class.isAssignableFrom(contextClass)) {
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogError)
					.message(
							"Context Wrapper "
									+ className
									+ " is not extending the WrapperClientContext!")
					.build().code(GateException.CODE_GENERIC_INVALID_CLASS);
		}

		try {
			@SuppressWarnings("unused")
			Constructor<? extends WrapperClientContext> constructor = (Constructor<? extends WrapperClientContext>) contextClass
					.getConstructor(IClientContext.class);
		} catch (NoSuchMethodException | SecurityException e) {
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogAndRelease)
					.message(
							"Context Wrapper "
									+ className
									+ " is not defining the constructor with IClienContext.class parameter!")
					.build().code(GateException.CODE_GENERIC_INVALID_CLASS);
		}

		contextWrapper = contextClass;
		if(log.isInfoEnabled())
			log.info("Context Wrapper " + className + " is registered !");
	}

	/**
	 * Nullifies the contextWrapper.
	 */
	public static void unregisterContextWrapper() {
		contextWrapper = null;
		if(log.isInfoEnabled()) log.info("Default context wrapper is unregistered!");
	}

	/**
	 * Gives the registered context wrapper
	 * 
	 * @return the registered context wrapper
	 */
	public static Class<? extends WrapperClientContext> getContextWrapper() {
		return contextWrapper;
	}

	/*
	 * Locators related
	 */
	/**
	 * Registers a Service Locator
	 * 
	 * @param className
	 *            locator class name
	 * @throws GateException
	 */
	@SuppressWarnings("unchecked")
	public static boolean registerLocator(String className)
			throws GateException {
		nullArgumentsCheck(className);
		// Create instance
		Class<? extends ILocator> locatorClass  = null;
		try {
			locatorClass =  (Class<? extends ILocator>) Class.forName(className);
		} catch (Exception e) {
			throw ExceptionBuilder
					.with(GateException.class, e)
					.execute(cmdLogError)
					.message(
							"An error occured while registering a locator!"
									+ className).build()
					.code(GateException.CODE_REGISTRY_UNEXPECTED);
		}
		return registerLocator(locatorClass);
	}

	/**
	 * Registers a Service Locator
	 * 
	 * @param locatorClass
	 *            locator class
	 * @throws GateException
	 */
	public static boolean registerLocator(Class<? extends ILocator> locatorClass)
			throws GateException {
		nullArgumentsCheck(locatorClass);
		// Create instance
		ILocator locator = null;
		try {
			locator = locatorClass.newInstance();
			if (locator == null) {
				throw ExceptionBuilder
						.with(GateException.class)
						.execute(cmdLogError)
						.message(
								"Locator class not found in the classpath!"
										+ locatorClass.getName()).build()
						.code(GateException.CODE_GENERIC_CLASS_NOT_FOUND);
			}
		} catch (Exception e) {
			throw ExceptionBuilder
					.with(GateException.class, e)
					.execute(cmdLogError)
					.message(
							"An error occured while registering a locator!"
									+ locatorClass.getName()).build()
					.code(GateException.CODE_REGISTRY_UNEXPECTED);
		}
		return registerLocator(locator);
	}

	/**
	 * Registers a Service Locator
	 * 
	 * @param locator
	 *            locator instance
	 * @throws GateException
	 */
	public static boolean registerLocator(ILocator locator)
			throws GateException {
		nullArgumentsCheck(locator);
		// Create instance
		if (locators.get(locator.getName()) != null) {
			if (log.isWarnEnabled())
				log.warn("Locator " + locator.getName()
						+ " is already registered!");
			return true;
		}
		locators.put(locator.getName(), locator);
		if (locators.get(locator.getName()).equals(locator)) {
			if (log.isInfoEnabled()) {
				log.info("Locator " + locator.getClass().getName()
						+ " registered as " + locator.getName());
			}
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Tells if there is locator registered with the given name
	 * 
	 * @param name
	 *            name of the locator
	 * @return true if the quetioned locator is registered
	 */
	public static boolean hasServiceLocator(String name) {
		return name == null ? false : locators.containsKey(name);
	}

	/**
	 * Gives the set of the registered locator names
	 * 
	 * @return set of registered locator names
	 */
	public static Set<String> getLocatorNames() {
		return locators.keySet();
	}

	/**
	 * Gives the instance of the registered locator
	 * 
	 * @param name
	 *            name of the locator
	 * @return the registered locator
	 */
	static ILocator findLocator(String name) {
		return locators.get(name);
	}

	/*
	 * Services Related
	 */
	/**
	 * Registers a service definition
	 * 
	 * @param def service definition
	 * @throws ExAlreadyRegistered
	 * @throws ServiceRegistrationException
	 */
	public static void registerService(ServiceDefinition def)
			throws GateException {
		// Validations
		nullArgumentsCheck(def);

		lockState();
		if (serviceDefinitions.get(def.getName()) != null) {
			releaseState();
			log.warn(def.getName() + " service is already registered!");
			return;
		}

		ServiceDefinition serviceDef = def.clone();

		ILocator loc = locators.get(serviceDef.getLocator());
		boolean locatorExist = true;
		if (loc == null) {
			locatorExist =false;
			log.warn("The locator ("
									+ serviceDef.getLocator() + ") that bounded to the service defininition ("
									+ serviceDef.getName() + ") is not exist yet!");
		}

		if (locatorExist && !serviceDef.isLazy()) {
			try {
				if (serviceDef.getServiceScope().equals(
						ServiceDefinition.ServiceScope.SINGLETON)) {
					Object si = loc.getComponentInstance(serviceDef);
					componentSingletons.put(serviceDef.getName(), si);
				}
			} catch (Exception e) {
				throw ExceptionBuilder.with(GateException.class, e)
						.execute(cmdLogAndRelease).message(e.getMessage())
						.build()
						.code(GateException.CODE_COMPONENT_INITIALIZATION);
			}
		}

		serviceDefinitions.put(serviceDef.getName(), serviceDef);

		log.info("Service " + serviceDef.getName() + " is registered ");

		releaseState();
	}

	/**
	 * Builds a service definition from the given {@link GateService}
	 * annotated interface and registers it
	 * 
	 * @param clazz
	 *            an interface that annotated with {@link GateService} annotation
	 * @throws GateException
	 */
	public static void registerService(Class<?> clazz) throws GateException {

		if (clazz == null) {
			throw new IllegalArgumentException("Class is null");
		}

		if (!clazz.isAnnotationPresent(GateService.class)) {
			throw new IllegalArgumentException(
					"Class is not annotated with @GateService");
		}
		ServiceDefinition sDef = new ServiceDefinition(clazz);

		if (serviceDefinitions.get(sDef.getName()) != null) {
			log.warn(sDef.getName() + " service is already registered!");
			return;
		}

		Method[] methods = clazz.getMethods();
		boolean opAdded = false;
		for (Method method : methods) {
			if (!method.isAnnotationPresent(GateOperation.class))
				continue;
			OperationDefinition opDef = new OperationDefinition(method);
			sDef.addOperation(opDef);
			opAdded = true;
		}
		if (!opAdded) {
			// Add all methods
			for (Method method : methods) {
				if (!Modifier.isPublic(method.getModifiers()))
					continue;
				OperationDefinition opDef = new OperationDefinition(method);
				sDef.addOperation(opDef);
			}
		}
		registerService(sDef);
	}

	/**
	 * Gives the map of the registered service definitions.
	 * 
	 * @return map of registered service definitions
	 * @throws ExComponentInitialization
	 * @throws ExComponentNotFound
	 */
	public static Map<String, ServiceDefinition> getServiceDefinitions(
			IClientContext context) {
		Map<String, ServiceDefinition> rmap = new HashMap<>();
		for (String key : serviceDefinitions.keySet()) {
			try {
				ServiceDefinition service = getServiceDefinition(key, context,false,true);
				if(service != null) rmap.put(key, service);
			} catch (GateException e) {
			}
		}
		return rmap;
	}

	/**
	 * Gives a service definiton by it's name
	 * 
	 * @throws GateException
	 */
	public static ServiceDefinition getServiceDefinition(String serviceName,
			IClientContext context) throws GateException {
		return getServiceDefinition(serviceName, context, false,false);
	}

	private static ServiceDefinition getServiceDefinition(String serviceName,
			IClientContext context, boolean toUse,boolean slient) throws GateException {
		ServiceDefinition def = getServiceDefinitionWithoutSecurity(
				serviceName, context);
		if(! checkServiceAccessSecurity(def, context, toUse,slient))	 return null;
		ServiceDefinition ret = def.clone();
		if (def.getOperations().size() > 0) {
			for (OperationDefinition opDef : def.getOperations()) {
				try {
					if(! checkOperationAccessSecurity(def, opDef, context, toUse,slient)){
						ret.removeOperation(opDef);
					}
				} catch (GateException e) {
					ret.removeOperation(opDef);
				}
			}
		}
		return ret;
	}

	private static ServiceDefinition getServiceDefinitionWithoutSecurity(
			String serviceName, IClientContext context) throws GateException {
		ServiceDefinition def = serviceDefinitions.get(serviceName);
		if (def == null) {
			throw ExceptionBuilder.with(GateException.class)
					.fill("%1! service not found in the registry", serviceName)
					.execute(cmdLogDebug).build()
					.code(GateException.CODE_REGISTRY_SERVICE_NOT_FOUND);
		}

		if (def.getOperations().size() > 0) {
			return def.clone();
		}
		// We need to generate op definition from service instance's methods
		ReentrantLock lock = serviceLock.get(serviceName);
		if (lock == null) {
			lock = new ReentrantLock();
			ReentrantLock existLock = serviceLock
					.putIfAbsent(serviceName, lock);
			if (existLock != null)
				lock = existLock;
		}
		// We like to suspend giving service information until we build it
		lock.lock();
		ServiceDefinition ret = null;
		// A parallel thread may already build it
		try {
			def = serviceDefinitions.get(serviceName);
			ret = def.clone();
			if (def.getOperations().size() == 0) {
				Object instance = getServiceInstance(def, context);

				Method lOBMethods[] = instance.getClass().getDeclaredMethods();
				for (Method method : lOBMethods) {
					String opName = method.getName();
					if (def.getOperation(opName) == null) {
						OperationDefinition opDef = new OperationDefinition(
								opName);
						def.addOperation(opDef);
						ret.addOperation(opDef);
					}
				}

			}
		} catch (Exception e) {
			// Don't need to handle exception , just to need release the lock
			throw e;
		} finally {
			// release lock
			lock.unlock();
			serviceLock.remove(serviceName);
		}
		return ret;
	}

	private static Object getServiceInstance(ServiceDefinition sDef,
			IClientContext clientContext) throws GateException {

		String cmpName = sDef.getComponentTarget();
		// Find the locator and attach the service def
		ILocator loc = locators.get(sDef.getLocator());
		if (loc == null) {
			throw ExceptionBuilder.with(GateException.class)
					.message("Locator " + sDef.getLocator() + " is not found!")
					.execute(cmdLogError).build()
					.code(GateException.CODE_COMPONENT_INITIALIZATION);
		}

		Object component = null;
		if (sDef.getServiceScope().equals(
				ServiceDefinition.ServiceScope.SINGLETON)) {
			component = componentSingletons.get(cmpName);
			if (component == null) {
				component = loc.getComponentInstance(sDef);
			}
			if (component != null)
				componentSingletons.put(cmpName, component);
		} else if (sDef.getServiceScope().equals(
				ServiceDefinition.ServiceScope.SESSION)
				&& clientContext != null) {
			IDataStage session = clientContext.getDataStage(Scope.SESSION);
			if (session == null) {
				component = loc.getComponentInstance(sDef);
			} else {
				component = session.getAttribute(SESSION_KEY_PREFIX + cmpName);

				if (component == null) {
					component = loc.getComponentInstance(sDef);
				}
				if (component != null) {
					session.setAttribute(SESSION_KEY_PREFIX + cmpName,
							component);
				}
			}
		} else {
			component = loc.getComponentInstance(sDef);
		}

		return component;
	}

	/**
	 * Finds operation name matching object methods that belongs to a service
	 * component instance
	 * 
	 * @param serviceName
	 *            Name of the service
	 * @param operationName
	 *            Name of the operation
	 * @param clientContext
	 *            The caller context
	 * @return list of name mathcing methods
	 * @throws GateException
	 */
	public static List<Method> findNameMatchingMethods(String serviceName,
			String operationName, IClientContext clientContext)
			throws GateException {

		// find service def
		ServiceDefinition sDef = getServiceDefinition(serviceName,
				clientContext);

		Object instance = getServiceInstance(sDef, clientContext);

		OperationDefinition opDef = sDef.getOperation(operationName);
		if (opDef == null) {
			throw ExceptionBuilder
					.with(GateException.class)
					.fill("Operation %1! not found in %2!", operationName,
							serviceName).execute(cmdLogError).build()
					.code(GateException.CODE_REGISTRY_OPERATION_NOT_FOUND);
		}
		return findMethods(instance, opDef.getMethodTarget());
	}

	/*
	 * Scan and register
	 */
	
	/**
	 * Scans given package names and registers Gate related locators, interceptors and services 
	 * 
	 * @param packages name of the package patterns to be scanned
	 */
	public static void scanAndRegister(String... packages) {
		if(packages==null) return;
		List<String> packageNames = new ArrayList<String>(packages.length);
		for (String name : packages) {
			packageNames.add(name);
		}
		scanAndRegister(packageNames,null);
	}
	
	/**
	 * Scans given package names and registers Gate related locators, interceptors and services 
	 * 
	 * @param packageNames name of the package patterns to be scanned
	 */
	public static void scanAndRegister(List<String> packageNames) {
		scanAndRegister(packageNames,null);
	}
	
	/**
	 * 
	 * Scans given package names and registers Gate related locators, interceptors and services 
	 * 
	 * @param packageNames name of the package patterns to be scanned
	 * @param excludedPackageNames name of the package patterns to be excluded
	 */
	@SuppressWarnings("unchecked")
	public static void scanAndRegister(List<String> packageNames,List<String> excludedPackageNames) {
		if(packageNames == null || packageNames.size() == 0) return;
		if(log.isInfoEnabled()){
			log.info("Scanning and registering packages "+packageNames+". (Exlusions:"+excludedPackageNames+")");
		}
		//Package names 
		ResourceFilter packages = new ResourceFilter();
		for (String packageName : packageNames) {
			packages.packageName(packageName);
		}
		//Excluded package names
		ResourceFilter epackages = new ResourceFilter();
		if(excludedPackageNames != null){
			for (String packageName : excludedPackageNames) {
				epackages.packageName("!"+packageName);
			}	
		}
		CombinedFilter packagesFilter = new CombinedFilter();
		packagesFilter.appendFilter(packages).appendFilter(epackages).combineWithAnd();
		//Gate Types
		CombinedFilter typesFilter = new CombinedFilter();
		// Locators
		typesFilter.appendFilter(new ClassFilter()
				.interfaceClass(ILocator.class).classOnly());
		// GateService annotated Interfaces
		typesFilter.appendFilter(new ClassFilter().annotation(
				GateService.class).interfaceOnly());
		// Advice and Listeners
		typesFilter.appendFilter(new ClassFilter()
				.interfaceClass(IExecutionAdvice.class)
				.annotation(GateServiceTarget.class)
				.annotation(GateServiceTargets.class)
				.joinAnnotationsWithOr().joinInterfacesWithOr().classOnly());
		// Filters
		typesFilter.appendFilter(new ClassFilter()
				.interfaceClass(IServiceFilter.class).classOnly());
		// Scan packages
		List<Class<?>> classes = CPScanner.scanClasses(packagesFilter,typesFilter);
		// Register founds
		for (Class<?> clazz : classes) {
			try {
				if (ILocator.class.isAssignableFrom(clazz)) {
					registerLocator((Class<? extends ILocator>) clazz);
				}else if (IServiceFilter.class.isAssignableFrom(clazz)) {
					registerFilter((Class<? extends IServiceFilter>) clazz);
				}else if(IExecutionAdvice.class.isAssignableFrom(clazz)){
					registerAdvice((IExecutionAdvice) clazz.newInstance());
				}else{
					registerService(clazz);
				}
			} catch (Exception e) {
				ExceptionCatcher.trap(e).execute(cmdLogError).process();
			}
		}
	}

	/*
	 * Others
	 */

	/**
	 * Gives executable target for the given service and operation name
	 * 
	 * @param clientContext
	 *            The caller context
	 * @param serviceName
	 *            The name of the service
	 * @param operationName
	 *            The operation name
	 * @param parameters
	 *            parameter object
	 * @return ExecutableTarget executable method
	 * @throws GateException
	 */
	static ExecutionTarget getExecutionTarget(IClientContext clientContext,
			String serviceName, String operationName, Object... parameters)
			throws GateException {
		Class<?>[] classes = parameters == null ? null
				: new Class<?>[parameters.length];
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				classes[i] = parameters[i] == null ? null : parameters[i]
						.getClass();
			}
		}
		return getExecutionTarget(clientContext, serviceName, operationName,
				classes);
	}

	/**
	 * Gives an executable target for the given parameter types
	 * 
	 * @param clientContext
	 * @param serviceName
	 * @param operationName
	 * @param parameterTypes
	 * @return execution target
	 * @throws GateException
	 */
	static ExecutionTarget getExecutionTarget(IClientContext clientContext,
			String serviceName, String operationName,
			Class<?>... parameterTypes) throws GateException {

		// find service def
		ServiceDefinition sDef = getServiceDefinitionWithoutSecurity(
				serviceName, clientContext);
		checkServiceAccessSecurity(sDef, clientContext, true,false);
		// Check if callable
		if (!sDef.isRemoteAccessEnabled() && clientContext.isLocal()) {
			throw ExceptionBuilder
					.with(GateException.class)
					.fill("Service %1! is only allowed to use for local calls",
							serviceName).execute(cmdLogError).build()
					.code(GateException.CODE_SERVICE_LOCAL_USE_ONLY);
		}

		// find operation definition
		OperationDefinition opDef = sDef.getOperation(operationName);
		if (opDef == null) {
			throw ExceptionBuilder
					.with(GateException.class)
					.fill("%1!.%2! operation is not found", serviceName,
							operationName).execute(cmdLogError).build()
					.code(GateException.CODE_REGISTRY_OPERATION_NOT_FOUND);
		}
		checkOperationAccessSecurity(sDef, opDef, clientContext, true,false);
		// Get instance
		Object instance = getServiceInstance(sDef, clientContext);
		if (instance == null) {
			throw ExceptionBuilder
					.with(GateException.class)
					.fill("%1!.%2! initialization returned null!", serviceName,
							operationName).execute(cmdLogError).build()
					.code(GateException.CODE_COMPONENT_INITIALIZATION);
		}
		// Find the method
		Method method = findMethod(instance, opDef.getMethodTarget(),
				parameterTypes, false, clientContext);
		if (method == null) {
			throw ExceptionBuilder
					.with(GateException.class)
					.fill("%1!.%2! matching method not found !", serviceName,
							operationName).execute(cmdLogError).build()
					.code(GateException.CODE_REGISTRY_OPERATION_NOT_FOUND);
		}
		return new ExecutionTarget(sDef, opDef, instance, method);
	}

	/**
	 * Searches the given method with the given parameters in the given class
	 * instance
	 * 
	 * <p>
	 * Sample : <code> TestClass lOBTest = new TestClass();
	 * 				  Method lOBMethod = DynaCall.findMethod(lOBTest,
	 * 						"testMethod",lOBParameterTypes);
	 * 				   </code>
	 * </p>
	 * 
	 * @param instance
	 * @param pOBMethod
	 * @param parameters
	 * @return returns Method object if found any
	 */
	private static Method findMethod(Object serviceInstance,
			String operationName, Class<?>[] parameterTypes,
			boolean firstCycle, IClientContext context) {
		Method[] lOBMethods = serviceInstance.getClass().getMethods();

		if (lOBMethods == null)
			return null;

		for (Method method : lOBMethods) {
			if (!method.getName().equals(operationName))
				continue;
			// methodu bulduk
			Class<?>[] lOBParameterTypes = method.getParameterTypes();
			// parametre say�lar� da ayn�
			if (lOBParameterTypes.length == (parameterTypes == null ? 0
					: parameterTypes.length)) {
				// t�m parametreler convertible'mi ?
				int lincnt = 0;
				for (lincnt = 0; lincnt < lOBParameterTypes.length; lincnt++) {
					if (lOBParameterTypes[lincnt] == null)
						continue;
					// Parametre primitive type ve Object ile ayn? tiptemi
					if (lOBParameterTypes[lincnt].isPrimitive()) {
						if (parameterTypes[lincnt] == null
								|| !isSameType(
										lOBParameterTypes[lincnt].getName(),
										parameterTypes[lincnt].getName())) {
							break;
						}
					} else if (parameterTypes[lincnt] != null
							&& firstCycle
							&& !lOBParameterTypes[lincnt]
									.equals(parameterTypes[lincnt])) {
						break;
					} else if (parameterTypes[lincnt] != null
							&& !firstCycle
							&& !lOBParameterTypes[lincnt]
									.isAssignableFrom(parameterTypes[lincnt])) {
						break;
					}
				}
				// do�ru metodu bulduk!
				if (lincnt == lOBParameterTypes.length) {
					return method;
				}
			}
		}
		if (!firstCycle) {
			return findMethod(serviceInstance, operationName, parameterTypes,
					false, context);
		} else {
			return null;
		}

	}

	private static boolean isSameType(String pstPrimitiveType,
			String pstObjectType) {
		if ((pstPrimitiveType.compareTo(pstObjectType) == 0)
				|| (pstPrimitiveType.compareTo("short") == 0 && pstObjectType
						.compareTo("java.lang.Short") == 0)
				|| (pstPrimitiveType.compareTo("byte") == 0 && pstObjectType
						.compareTo("java.lang.Byte") == 0)
				|| (pstPrimitiveType.compareTo("int") == 0 && pstObjectType
						.compareTo("java.lang.Integer") == 0)
				|| (pstPrimitiveType.compareTo("long") == 0 && pstObjectType
						.compareTo("java.lang.Long") == 0)
				|| (pstPrimitiveType.compareTo("double") == 0 && pstObjectType
						.compareTo("java.lang.Double") == 0)
				|| (pstPrimitiveType.compareTo("float") == 0 && pstObjectType
						.compareTo("java.lang.Float") == 0)
				|| (pstPrimitiveType.compareTo("char") == 0 && pstObjectType
						.compareTo("java.lang.Char") == 0)
				|| (pstPrimitiveType.compareTo("boolean") == 0 && pstObjectType
						.compareTo("java.lang.Boolean") == 0)) {
			return true;
		} else {
			return false;
		}

	}

	private static List<Method> findMethods(Object instance, String methodName) {
		List<Method> foundMethods = new ArrayList<>();
		Method lOBMethods[] = instance.getClass().getDeclaredMethods();
		for (int linI = 0; linI < lOBMethods.length; linI++) {
			// methodu bulduk
			if (lOBMethods[linI].getName().compareTo(methodName) == 0) {
				foundMethods.add(lOBMethods[linI]);
			}
		}
		return foundMethods;
	}

	private static boolean checkOperationAccessSecurity(
			ServiceDefinition serviceDef, OperationDefinition opDef,
			IClientContext context, boolean checkToUse, boolean slient) throws GateException {
		// Check Protocol
		if (opDef.isSecureProtocolRequiredToInspect()
				&& context.isProtocolSecure() == false) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.message(
							"Protocol security is required to inspect  "
									+ opDef.getName() + " operation on "
									+ serviceDef.getName() + " service")
					.execute(cmdLogDebug)
					.build()
					.code(GateException.CODE_SECURITY_PROTOCOL_SECURITY_REQUIRED);
		}
		if (checkToUse && opDef.isSecureProtocolRequiredToInspect()
				&& context.isProtocolSecure() == false) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.message(
							"Protocol security is required to use  "
									+ opDef.getName() + " operation on "
									+ serviceDef.getName() + " service")
					.execute(cmdLogDebug)
					.build()
					.code(GateException.CODE_SECURITY_PROTOCOL_SECURITY_REQUIRED);
		}

		// check authentication
		if (opDef.isAuthenticationRequiredToInspect()
				&& context.getClientUser() == null) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.message(
							"Authentication is required to inspect  "
									+ opDef.getName() + " operation on "
									+ serviceDef.getName() + " service")
					.execute(cmdLogDebug).build()
					.code(GateException.CODE_SECURITY_AUTHENTICATION_REQUIRED);
		}
		if (checkToUse && opDef.isAuthenticationRequiredToUse()
				&& context.getClientUser() == null) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.message(
							"Authentication is required to use  "
									+ serviceDef.getName() + " service")
					.execute(cmdLogError).build()
					.code(GateException.CODE_SECURITY_AUTHENTICATION_REQUIRED);
		}

		// check authorization
		boolean authorized = opDef.getAuthorizedRolesToInspect().size() <= 0;
		if (!authorized) {
			for (String role : opDef.getAuthorizedRolesToInspect()) {
				authorized = context.isUserInRole(role);
				if (authorized)
					break;
			}
			if (!authorized) {
				if(slient) return false;
				throw ExceptionBuilder
						.with(GateException.class)
						.message(
								"You're not authorized to inspect "
										+ opDef.getName() + " operation on "
										+ serviceDef.getName() + " service")
						.execute(cmdLogDebug).build()
						.code(GateException.CODE_SECURITY_NOT_AUTHORIZED);
			}
		}
		if (checkToUse) {
			authorized = opDef.getAuthorizedRolesToUse().size() <= 0;
			if (!authorized) {
				for (String role : opDef.getAuthorizedRolesToUse()) {
					authorized = context.isUserInRole(role);
					if (authorized)
						break;
				}
				if (!authorized) {
					if(slient) return false;
					throw ExceptionBuilder
							.with(GateException.class)
							.execute(cmdLogDebug)
							.message(
									"You're not authorized to use "
											+ opDef.getName()
											+ " operation on "
											+ serviceDef.getName() + " service")
							.build()
							.code(GateException.CODE_SECURITY_NOT_AUTHORIZED);
				}
			}
		}
		return true;
	}

	private static boolean checkServiceAccessSecurity(
			ServiceDefinition serviceDef, IClientContext context,
			boolean checkToUse,boolean slient) throws GateException {

		// Check Protocol
		if (serviceDef.isSecureProtocolRequiredToInspect()
				&& context.isProtocolSecure() == false) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogDebug)
					.message(
							"Protocol security is required to inspect  "
									+ serviceDef.getName() + " service")
					.build()
					.code(GateException.CODE_SECURITY_PROTOCOL_SECURITY_REQUIRED);

		}
		if (checkToUse && serviceDef.isSecureProtocolRequiredToInspect()
				&& context.isProtocolSecure() == false) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogDebug)
					.message(
							"Protocol security is required to use  "
									+ serviceDef.getName() + " service")
					.build()
					.code(GateException.CODE_SECURITY_PROTOCOL_SECURITY_REQUIRED);
		}

		// check authentication & authorization at the service level
		if (serviceDef.isAuthenticationRequiredToInspect()
				&& context.getClientUser() == null) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogDebug)
					.message(
							"Authentication is required to inspect  "
									+ serviceDef.getName() + " service")
					.build()
					.code(GateException.CODE_SECURITY_AUTHENTICATION_REQUIRED);
		}
		if (checkToUse && serviceDef.isAuthenticationRequiredToUse()
				&& context.getClientUser() == null) {
			if(slient) return false;
			throw ExceptionBuilder
					.with(GateException.class)
					.execute(cmdLogDebug)
					.message(
							"Authentication is required to use  "
									+ serviceDef.getName() + " service")
					.build()
					.code(GateException.CODE_SECURITY_PROTOCOL_SECURITY_REQUIRED);
		}

		// check authorization at the service level
		boolean authorized = serviceDef.getAuthorizedRolesToInspect().size() <= 0;
		if (!authorized) {
			for (String role : serviceDef.getAuthorizedRolesToInspect()) {
				authorized = context.isUserInRole(role);
				if (authorized)
					break;
			}
			if (!authorized) {
				if(slient) return false;
				throw ExceptionBuilder
						.with(GateException.class)
						.execute(cmdLogDebug)
						.message(
								"You're not authorized to inspect "
										+ serviceDef.getName() + " service")
						.build()
						.code(GateException.CODE_SECURITY_NOT_AUTHORIZED);
			}
		}
		if (checkToUse) {
			authorized = serviceDef.getAuthorizedRolesToUse().size() <= 0;
			if (!authorized) {
				for (String role : serviceDef.getAuthorizedRolesToUse()) {
					authorized = context.isUserInRole(role);
					if (authorized)
						break;
				}
				if (!authorized) {
					if(slient) return false;
					throw ExceptionBuilder
							.with(GateException.class)
							.execute(cmdLogDebug)
							.message(
									"You're not authorized to use "
											+ serviceDef.getName() + " service")
							.build()
							.code(GateException.CODE_SECURITY_NOT_AUTHORIZED);
				}
			}
		}
		return true;
	}

	/*
	 * Helpers
	 */
	private static void nullArgumentsCheck(Object... parameters)
			throws IllegalArgumentException {
		int inx = 0;
		for (Object param : parameters) {
			inx++;
			if (param == null) {
				ExceptionBuilder.with(IllegalArgumentException.class)
						.message("Parameter " + inx + " is null!")
						.execute(cmdLogError).throwIt();
			}
		}
	}

	/*
	 * Locking State Related Methods
	 */
	private static void lockState() {
		lock.lock();
	}

	private static void releaseState() {
		if (lock.isHeldByCurrentThread())
			lock.unlock();
	}

}
