package net.sf.corn.gate.jsonrpc;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import net.sf.corn.converter.ConversionException;
import net.sf.corn.converter.ParsingException;
import net.sf.corn.converter.json.JsTypeComplex;
import net.sf.corn.converter.json.JsTypeList;
import net.sf.corn.converter.json.JsTypeObject;
import net.sf.corn.converter.json.JsTypeSimple;
import net.sf.corn.converter.json.JsonConverter;
import net.sf.corn.converter.json.JsonStringParser;

import org.springframework.beans.ConversionNotSupportedException;

/**
 * This class represents service requests to JSON-RPC services. JsonRpc client
 * and dispatcher uses this class internally. 
 * The following Java Script class shows a sample request message structure.
 * <code><br>
 *  {<br>
 *   "jsonrpc":"2.0"<br>
 *   "id":"64f4f4b0-79ee-45f0-9b47-77836bc74ce8",<br>
 *   "method":"service.operation",<br>
 *   "header":{ <br>
 *     "headerField1": "headerVal",<br>
 *     "headerField2": "headerVal",<br>
 *  },<br>
 *   
 *   "parameters":["param1","param2"]<br> 
 *   }<br>
 * </code><br>
 * <h2>Message Parts</h2> 
 * <h3>jsonrpc</h3>
 * "2.0" value is standard for JSON-RPC 2.0 messages 
 * <h3>id</h3> 
 * This part should contain a unique message
 * identifier for the message correlation. Results of operations will be sent to
 * the clients by using same identifier. This would help clients to correlate
 * their responses from server. 
 * <h3>method</h3> 
 * The remote service name and operation name that seperated with "." character like "service.operation".
 * <h3>header</h3> 
 * The header part of the message can contain anything that you want
 * to send to the server beside parameters.The header part is not accessible by
 * services. Interceptors can access header fields to apply generic policies to
 * service calls. For example you can send additional security credentials to
 * create your own security. 
 * <h3>parameters</h3> 
 * This part contains the target operation's parameters as an array of values.
 * 
 * 
 * @author Serhat Dirik
 * 
 */
class JsonRpcRequest implements Serializable{

	private static final long serialVersionUID = -312180336124254589L;
	public static final String METHOD_ATTRIBUTE = "method";
	public static final String PARAMETERS_ATTRIBUTE = "params";
	public static final String REQUESTID_ATTRIBUTE = "id";
	public static final String JSONRPC_ATTRIBUTE = "jsonrpc";
	public static final String HEADER_ATTRIBUTE = "header";
	private JsTypeComplex enclosedRequest;
	private transient JsonConverter converter = new JsonConverter();
	/**
	 * Creates new empty message
	 */
	public JsonRpcRequest() {
		enclosedRequest = new JsTypeComplex();
		setJsonrpcVersion("2.0");
	}

	/**
	 * Creates new empty message
	 * @throws ParsingException 
	 * @throws JsonException 
	 * @throws ConversionNotSupportedException 
	 * @throws ConversionException 
	 */
	public JsonRpcRequest(String jsonString) throws ParsingException, ConversionException  {
		JsTypeComplex json_req = (JsTypeComplex) JsonStringParser.parseJsonString(jsonString);
		setEnclosedRequest(json_req);
	}
	
 
	/**
	 * Resets all parts of the message and assigns the from the given message
	 * parts
	 * 
	 * @param jsonReq
	 * @throws ConversionException
	 * @throws ConversionNotSupportedException
	 */
	private void setEnclosedRequest(JsTypeComplex jsonReq) throws ConversionException {
		JsTypeObject jsObj = null;
		if (jsonReq != null) { //Validate required parts!
			for(String key: jsonReq.keys()){
				jsObj = jsonReq.get(key);
				if(key.equals(METHOD_ATTRIBUTE)){
					String method = jsObj != null ? jsObj.toStringValue() : null;
					if (method == null) {
						throw new ConversionException("method not found");
					}
					String[] methodParts = method.split("\\.");
					if(methodParts.length != 2){
						throw new ConversionException("method must be formatted as service.operation!");
					}
				}else if(key.equals(HEADER_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isComplex()) {
						throw new ConversionException("header must be must be a complex one!");
					}
					
				}else if(key.equals(PARAMETERS_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isList()) {
						throw new ConversionException("params must be a list!");
					}
				}else if(key.equals(JSONRPC_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isSimple()) {
						throw new ConversionException("jsonrpc must be a simple!");
					}
				}else if(key.equals(REQUESTID_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isSimple()) {
						throw new ConversionException("id must be a simple!");
					}
				}else{
					throw new ConversionException("Message contains illegal attribute name "+key);
				}
			}
			this.enclosedRequest = jsonReq;
		} else {
			throw new ConversionException("Empty request in");
		}

	}

	/**
	 * Assigns the given parameter to the service name. The service name and
	 * message name will be merged in method part of the message in dotted form
	 * like "servicename.operationname".
	 * 
	 * @param serviceName
	 * @throws JSONException
	 */
	public void setServiceName(String serviceName) {
		if(serviceName == null){
			throw new IllegalArgumentException("Null serviceName in!");
		}
		String method = serviceName;
		if(getOperationName() != null){
			method += "."+getOperationName();
		}
		try {
			enclosedRequest.putObject(METHOD_ATTRIBUTE, new JsTypeSimple(JsTypeObject.quote(method),null));
		} catch (ConversionException e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	/**
	 * Assigns the given parameter to the method name. The method name and
	 * service name will be merged in method part of the message in dotted form
	 * like "servicename.operationName".
	 * 
	 * @param opName operation Name
	 * @throws JSONException
	 */
	public void setOperationName(String opName) {
		if(opName == null){
			throw new IllegalArgumentException("Null opName in!");
		}
		String method = null;
		if(getServiceName() != null){
			method = getServiceName()+"."+opName;
		}else{
			method =  "."+opName;
		}
		try {
			enclosedRequest.putObject(METHOD_ATTRIBUTE, new JsTypeSimple(JsTypeObject.quote(method),null));
		} catch (ConversionException e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	private void setJsonrpcVersion(String jsonrpcVersion) {
		try {
			enclosedRequest.putObject(JSONRPC_ATTRIBUTE, new JsTypeSimple(JsTypeObject.quote(jsonrpcVersion),null));
		} catch (ConversionException e) {
			throw new IllegalArgumentException(e);
		}
	}
 
	

	/**
	 * Sets given object array as parameters
	 * 
	 * @param parameters
	 * @throws ConversionNotSupportedException 
	 * @throws ConversionException 
	 * @throws JSONException
	 */
	public void setParameters(Object[] parameters) throws ConversionException, ConversionNotSupportedException  {
		JsTypeList jsArr = (JsTypeList) converter.toTargetType(parameters);
		setParameters(jsArr);
	}

	/**
	 * Sets given JSONArray as parameters
	 * @param parameters
	 * @throws JSONException
	 */
	public void setParameters(JsTypeList parameters){
		enclosedRequest.put(PARAMETERS_ATTRIBUTE, parameters);
	}

	/**
	 * Sets the unique request id for the message correlation
	 * 
	 * @param requestId
	 * @throws ConversionNotSupportedException 
	 * @throws ConversionException 
	 * @throws JSONException
	 */
	public void setRequestId(String requestId){
		enclosedRequest.put(REQUESTID_ATTRIBUTE,  new JsTypeSimple(JsTypeObject.quote(requestId),null));
	}

	/**
	 * Sets the header part of the message
	 * 
	 * @param header
	 * @throws JSONException
	 * @throws ConversionException
	 * @throws ConversionNotSupportedException
	 */
	public void setHeader(Map<String, Object> header) throws ConversionException, ConversionNotSupportedException{
		if (header == null){
			enclosedRequest.remove(HEADER_ATTRIBUTE);
			return;
		}
		JsTypeObject jsHdr = converter.toTargetType(header);
		enclosedRequest.put(HEADER_ATTRIBUTE, jsHdr);
	}

	/**
	 * Gives headers as unmodifiable collection
	 * 
	 * @return unmodifiable header part
	 */
	public Map<String, Object> getHeader() {
		if(! enclosedRequest.has(HEADER_ATTRIBUTE) || enclosedRequest.get(HEADER_ATTRIBUTE).isNull()) return new HashMap<String, Object>(0);;
		JsTypeComplex header = (JsTypeComplex) enclosedRequest.get(HEADER_ATTRIBUTE);
		try {
			@SuppressWarnings("unchecked")
			Map<String, Object> jheader = (Map<String, Object>) converter.toJava(HashMap.class, header);
			return jheader;
		} catch (ConversionException e) {
			e.printStackTrace();
			return new HashMap<String, Object>(0);
		}
	}

	/**
	 * Adds given field to the header
	 * 
	 * @param fieldName
	 * @param value
	 * @throws JSONException
	 * @throws ConversionException
	 * @throws ConversionNotSupportedException
	 */
	public void putToHeader(String fieldName, Object value) throws ConversionException, ConversionNotSupportedException{
		Map<String, Object> header = getHeader();
		if (header == null){
			header = new HashMap<String, Object>();
		}
		header.put(fieldName, value);
		enclosedRequest.put(HEADER_ATTRIBUTE, converter.toTargetType(header));
	}

	/**
	 * Removes the requested field form the header
	 * 
	 * @param fieldName
	 * @throws JSONException
	 * @throws ConversionException
	 * @throws ConversionNotSupportedException
	 */
	public void removeFromHeader(String fieldName) throws ConversionException, ConversionNotSupportedException{
		JsTypeComplex header = (JsTypeComplex) enclosedRequest.get(HEADER_ATTRIBUTE);
		if (header == null) return;
		header.remove(fieldName);
		enclosedRequest.put(HEADER_ATTRIBUTE, header);
	}

	public String getServiceName() {
		String method = null;
		if(enclosedRequest.has(METHOD_ATTRIBUTE)){
			method = enclosedRequest.get(METHOD_ATTRIBUTE).toStringValue();
		}else{
			return null;
		}
		if(method.indexOf('.') > 0){
			return method.substring(0,method.indexOf('.'));
		}else if(method.contains(".")){
			return null;
		}else{
			return method;
		}
	}

	public String getOperationName() {
		String method = null;
		if(enclosedRequest.has(METHOD_ATTRIBUTE)){
			method = enclosedRequest.get(METHOD_ATTRIBUTE).toStringValue();
		}else{
			return null;
		}
		if(method.indexOf('.') >= 0 && method.indexOf('.') < method.length()){
			return method.substring(method.indexOf('.') +1);
		}else{
			return null;
		}
	}

	public JsTypeList getParameters() {
		if(enclosedRequest.has(PARAMETERS_ATTRIBUTE) && ! enclosedRequest.get(PARAMETERS_ATTRIBUTE).isNull()){
			return (JsTypeList) enclosedRequest.get(PARAMETERS_ATTRIBUTE);
		}else{
			return new JsTypeList();
		}
	}

	public String getRequestId() {
		if(enclosedRequest.has(REQUESTID_ATTRIBUTE)){
			return   enclosedRequest.get(REQUESTID_ATTRIBUTE).toStringValue();
		}else{
			return null;
		}
	}

	public String getJsonrpcVersion() {
		if(enclosedRequest.has(JSONRPC_ATTRIBUTE)){
			return   enclosedRequest.get(JSONRPC_ATTRIBUTE).toStringValue();
		}else{
			return null;
		}
	}

	/**
	 * Converts message to JSONObject.
	 * 
	 * @return JSON Object
	 */
	public JsTypeObject toJSON() {
		return enclosedRequest;
	}

	/**
	 * Converts message to java script.
	 * 
	 * @return text in javascript notation
	 */
	public String toString() {
		return enclosedRequest == null ? null : enclosedRequest.toString();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((enclosedRequest == null) ? 0 : enclosedRequest.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		JsonRpcRequest other = (JsonRpcRequest) obj;
		if (enclosedRequest == null) {
			if (other.enclosedRequest != null)
				return false;
		} else if (!enclosedRequest.equals(other.enclosedRequest))
			return false;
		return true;
	}

	 

}
