package net.sf.corn.gate.locators;

import java.lang.reflect.Field;
import java.util.Hashtable;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import net.sf.corn.exception.ExceptionBuilder;
import net.sf.corn.exception.ExceptionCommand;
import net.sf.corn.gate.GateException;
import net.sf.corn.gate.ILocator;
import net.sf.corn.gate.ServiceDefinition;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * This class is a Component Locator implementation for EJB3 objects.This locator
 * access EJB instances by using JNDI, so the target attribute of components need
 * to show the JNDI name of the target EJB. <h2>Sample Configuration</h2>
 * &lt;Config&gt;<br>
 * &nbsp;&lt;Locators&gt;<br>
 * &nbsp;&nbsp;&lt;Locator class="net.sf.corn.gate.locator.EJB3Locator"/&gt;<br>
 * &nbsp;&lt;/Locators&gt;<br>
 * &nbsp;&lt;Services&gt;<br>
 * &nbsp;&nbsp;&lt;Service name="echoBean" componentTarget="EchoBean"
 * scope="singleton" locator="ejb3"&gt;<br>
 * &nbsp;&nbsp;&lt;/Service&gt;<br>
 * &nbsp;&lt;Services&gt;<br>
 * &lt;/Config&gt;<br>
* <p>
 * It's also possible to set additional attributes for JNDI context initialization such as context factory or jndi provider url.
 * Additional custom attributes are specified by {@link javax.naming.Context} classes constants. For example in order to set 
 * {@link javax.naming.Context} INITIAL_CONTEXT_FACTORY attribute, use contants string value as key which is 
 * "java.naming.factory.initial".
 * 
 * </p>
 * <h2>Sample Configuration</h2>
 * &lt;Config&gt;<br>
 *	&nbsp;&lt;Locators&gt;<br>
 * 		&nbsp;&nbsp;&lt;Locator class="net.sf.corn.gate.locator.EJB3Locator" /&gt;<br>
 *	&nbsp;&lt;/Locators&gt;<br>
 *	&nbsp;&lt;Services&gt;<br>
 *	  &nbsp;&nbsp;&lt;Service name="echoBean" componentTarget="GWTTests-EJB3-EJB3Session" scope="singleton" locator="ejb3"&gt;<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt!--INITIAL_CONTEXT_FACTORY--&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt;Property key="java.naming.factory.initial" value="weblogic.jndi.WLInitialContextFactory" /&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt!--PROVIDER_URL--&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt;Property key="java.naming.provider.url" value="t3://localhost:7001" /&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt!--SECURITY_PRINCIPAL--&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt;Property key="java.naming.security.principal" value="weblogic" /&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt!--SECURITY_CREDENTIALS--&gt<br>
 *			&nbsp;&nbsp;&nbsp;&nbsp;&lt;Property key="java.naming.security.credentials" value="weblogic1" /&gt<br>
 *		&nbsp;&nbsp;&lt;/Service&gt;<br>
 *	&nbsp;&lt;Services&gt;<br>
 * &lt;/Config&gt;<br>
 * @author <b>Serhat Dirik</b>
 * 
 */
public class EJB3Locator implements ILocator  {

	private String NAME = "ejb3";
	private static Logger log = LoggerFactory.getLogger(EJB3Locator.class);
	private InitialContext ctx = null;
	private static Set<String> contextAttributes= new CopyOnWriteArraySet<String>();
	private static ExceptionCommand cmdLog = new ExceptionCommand() {
		@Override
		public void execute(Exception e) {
         if(log.isErrorEnabled()) log.error(e.getMessage(),e);
			
		}
	};
	static{
		Field[] fields = Context.class.getFields();
		for (Field field : fields) {
			try {
				if(  field.getType().equals(String.class))  {
					String fieldVal = (String) 	field.get(null) ;
					contextAttributes.add(fieldVal);
				}
			} catch(Exception ex){
				
			}
		}
	}
	
	public EJB3Locator() {
		try {
			ctx = new InitialContext();
		} catch (NamingException e) {
			throw new RuntimeException(e);
		}
	}

	public Object getComponentInstance(ServiceDefinition compDef)
			throws GateException {
		InitialContext ejbCtx = ctx;

		Hashtable<String,String> env = new Hashtable<String,String>();
		for (String contextAttributeName  : contextAttributes) {
			String contextAttributeVal=compDef.getProperty(contextAttributeName);
			if(contextAttributeVal!=null)	env.put(contextAttributeName, contextAttributeVal);
		}
		//Try to initiate context with given parameters
		if(env.size() > 0){
			try {
				ejbCtx = new InitialContext( env );
			} catch (Exception e) {
				throw ExceptionBuilder.with(GateException.class,e).
						message("An exception occured while initialing JNDI context for the component "+
						compDef.getName()).execute(cmdLog).build().code(GateException.CODE_COMPONENT_INITIALIZATION);
			}
		}
		
		if (ejbCtx == null) {
			throw ExceptionBuilder.with(GateException.class).message("Naming Context is not ready").execute(cmdLog)
					.build().code(GateException.CODE_COMPONENT_INITIALIZATION);
		}
		
		try {
			Object component = ejbCtx.lookup(compDef.getComponentTarget());
			return component;
		} catch (Exception e) {
			throw ExceptionBuilder.with(GateException.class,e).message("An exception occured while looking EJB component "+compDef.getComponentTarget()).execute(cmdLog)
			.build().code(GateException.CODE_COMPONENT_INITIALIZATION);
		}
	}

	@Override
	public String getName() {
		return NAME;
	}

	@Override
	public String toString() {
		return getName()+" Locator";
	}
}
