package net.sf.corn.gate.web;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;

import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import net.sf.corn.exception.ExceptionBuilder;
import net.sf.corn.exception.ExceptionCommand;
import net.sf.corn.gate.XMLConfig;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class is designed to load configuration files at the beginning of Web
 * Container initialization. It gets "GateConfigFilesPattern" context
 * parameter and load files that are located under your <b>WEB-INF</b>
 * directory. You should add following lines to your web.xml file to activate
 * this listener.<br>
 * 
 * &nbsp;&lt;context-param&gt;<br>
 * &nbsp;&nbsp;&lt;description&gt;Config Files&lt;/description&gt;<br>
 * &nbsp;&nbsp;&lt;param-name&gt;GateConfigFilesPattern&lt;/param-name&gt;<br>
 * &nbsp;&nbsp;&lt;param-value&gt;Gate.*.xml&lt;/param-value&gt;<br>
 * &nbsp;&lt;/context-param&gt;<br>
 * &nbsp;&lt;listener&gt;<br>
 * &nbsp;&lt;description&gt;WebContextInitializer&lt;/description&gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;display-name&gt;WebContextInitializer&lt;/display-name&
 * gt;<br>
 * &nbsp;&nbsp;&nbsp;&lt;listener-class&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;org.gate.web.WebContextInitializer<br>
 * &nbsp;&nbsp;&nbsp;&lt;/listener-class&gt;<br>
 * &nbsp;&lt;/listener&gt;<br>
 * 
 * For some deployments it's necessary to explicitly specify configuration file directory in your web.xml file.Especially war deployments may fits to this case.
 * In that case please specify config file directory  as a context parameter as below.<br>
 * 
 * &nbsp;&lt;context-param&gt;<br>
 * &nbsp;&nbsp;&lt;description&gt;Config Files Directory&lt;/description&gt;<br>
 * &nbsp;&nbsp;&lt;param-name&gt;GateConfigPath&lt;/param-name&gt;<br>
 * &nbsp;&nbsp;&lt;param-value&gtc:/config&lt;/param-value&gt;<br>
 * &nbsp;&lt;/context-param&gt;<br>
 * 
 * @author Serhat Dirik
 * 
 */
public class GateContextInitializer implements ServletContextListener {

	private static Logger log = LoggerFactory.getLogger(GateContextInitializer.class);
	private static ExceptionCommand cmdLogError =new ExceptionCommand() {
		
		@Override
		public void execute(Exception e) {
			if(log.isDebugEnabled()) log.error(e.getMessage(),e);
		}
	};
	@Override
	public void contextInitialized(ServletContextEvent event) {
		log.info("Initialing Gate Context....");
		String configFile = event.getServletContext().getInitParameter(
				"GateConfigFilesPattern");
		log.info("GateConfigFilesPattern:" + configFile);
		if (configFile == null) {
			log.info("GateConfigFilesPattern context parameter not found in web.xml file!");
			return;
		}
		
		String configPath = event.getServletContext().getInitParameter(
				"GateConfigPath");
		log.info("GateConfigPath:"
				+ (configPath == null ? "WEB-INF" : configPath));

		if (configPath == null) {
			configPath = event.getServletContext().getRealPath("WEB-INF");
			if (configPath == null) {
				ExceptionBuilder.with(IllegalStateException.class).
					message("GateConfigConfigPath context parameter must be specified in web.xml file!").
					execute(cmdLogError).throwIt();
			}
		}
		
		File dir = new File(configPath);
		if(! dir.isDirectory()){
			ExceptionBuilder.with(IllegalStateException.class).
			message("GateConfigConfigPath context parameter (+"+configPath+") is not a directory!").
			execute(cmdLogError).throwIt();
		}
		
		boolean found = registerConfigFiles(dir, configFile);
		
		if(!found){
			ExceptionBuilder.with(IllegalStateException.class).
			fill("There is no config file found under %1! directory that matches %1! pattern that specified in web.xml file!",configPath).
			execute(cmdLogError).throwIt();
		}
		
		XMLConfig.processRegisteredXMLConfig();
	}

	private static boolean registerConfigFiles(File dir , String configFile){
		File[] files = dir.listFiles();
		boolean found = false;
		for (File file : files) {
			if (!file.isFile())
				continue;
			String name = file.getName().substring(
					file.getName().lastIndexOf(File.pathSeparator) + 1);
			if (file.isFile() && name.matches(configFile)) {
				try {
					XMLConfig
							.registerXMLConfig(new FileInputStream(file));
					found = true;
				} catch (FileNotFoundException e) {
					ExceptionBuilder.with(IllegalStateException.class)
					.fill("An error occured while trying to access %1! file that matches %2! pattern that specified in web.xml file!",file.getName(),configFile).
					execute(cmdLogError).throwIt();
				}
			}
		}
		return found;
	}
	
	
	@Override
	public void contextDestroyed(ServletContextEvent event) {

	}
}
